/* Radio.Unfix - radio.c - The Streamer */

#include "radio.h"

/* Configuration information */
struct Radio	radio;

/* Addresses and other stuff */
int			FileSocket = -1;

int iBaseTime;

/* Internal read function, which
* doesn't give up on EAGAIN errors
*/
int Read(int d, void *buf, size_t nbytes)
{
	int result;
	
	while (1)
	{
		result = read(d, buf, nbytes);
		/* Exit loop if result >=0 or errno is not EAGAIN */
		if (	result >= 0 ||
			errno != EAGAIN) break;
		dolog(LOG_ERR, "EAGAIN error in stream read\n");
	}
	return(result);
}

/* 
 * This routine sends out signals to processes that want them.
 *
 * Format of the signal file:
 * <cmd> <signal> <pid>
 *
 * Currently used commands:
 * titlechange - Send the signal when the title has changed
 *
 * eg:
 * titlechange 12 1234
 *
 * Send signal 12 (SIGUSR2) to process id 1234
 *
*/
void SendSignal(char *command)
{
	char buf[512], cmd[128];
	int signum, pid;

	if (!radio.c_SignalsFile) return;

	// Open the signals file
	FILE *f = fopen(radio.c_SignalsFile, "r");
	if (!f) return;

	while (fgets(buf, sizeof(buf), f))
	{
		if (3 != sscanf(
			buf,
			"%512s %d %d",
			cmd, &signum, &pid))
		{
			continue;
		}

		// Sanity checks
		if (	pid <= 0) continue;

		if (strcasecmp(command, cmd) == 0)
		{
			kill(pid, signum);
		}
	}
	fclose(f);
}

void SendTitle()
{
	char		oTitle[2048];
	unsigned int	i;
	int		ret;
	struct line	*line;

	SendSignal("titlechange");

	snprintf(oTitle, sizeof(oTitle),
		"x-audiocast-streamtitle: %s [%s]\nx-audiocast-streamurl: %s\n",
		radio.c_PlayingTitle, radio.c_StreamTitle, radio.c_URL);

	for (i=0;i<radio.interfaces;i++)
	{
		line = &radio.lines[i];
		ret = sendto(line->TitleSocket, oTitle, strlen(oTitle), 0,
			(struct sockaddr *)line->sTitleAddr->ai_addr, line->sTitleAddr->ai_addrlen);
		if (ret < 0)
		{
			dolog(LOG_ERR, "title sendto() returned: %m\n");
		}
	}
}

void SetPlayingTitle(char *song_id, char *artist, char *title)
{
	// Update the playlist
	query(	strlen(song_id),
		"INSERT INTO playlist (song_id, playlist_time) VALUES('%s', now())",
		song_id);

	// Update the number of times this song got played
	query(	strlen(song_id),
		"UPDATE songs SET song_played = song_played + 1 WHERE song_id = %s",
		song_id);

	// Set the playing title
	if (radio.c_PlayingTitle) free(radio.c_PlayingTitle);
	radio.c_PlayingTitle = malloc(strlen(artist) + strlen(title) + 4);
	if (!radio.c_PlayingTitle)
	{
		dolog(LOG_ERR, "Couldn't allocate memory for PlayingTitle\n");
		exit(-1);
	}
	memset(radio.c_PlayingTitle, 0, strlen(artist) + strlen(title) + 4);
	strcat(radio.c_PlayingTitle, artist);
	strcat(radio.c_PlayingTitle, " - ");
	strcat(radio.c_PlayingTitle, title);

	D(dolog(LOG_INFO, "Currently Playing: %s\n", radio.c_PlayingTitle);)
}

/*
 * Get the next stream
*/
void GetStream(int random)
{
	MYSQL_RES *res = NULL;
	MYSQL_ROW row = NULL;

	while (FileSocket == -1)
	{
		char *q = NULL;
		if (!random)
		{
			q = 
			"SELECT "\
			"COUNT(songs.song_id) AS numvotes, "\
			"songs.song_id, "\
			"song_filename, "\
			"song_artist, "\
			"song_title "\
			"FROM votes "\
			"INNER JOIN songs ON votes.song_id = songs.song_id "\
			"INNER JOIN pools ON songs.pool_id = pools.pool_id "\
			"WHERE pool_online = 'Y' AND vote_handled = 'N' "\
			"GROUP BY songs.song_id "\
			"ORDER BY numvotes DESC, song_artist, song_title "\
			"LIMIT 0,1";
		}
		else
		{
			q =
			"SELECT "\
			"0, "\
			"song_id, "\
			"song_filename, "\
			"song_artist, "\
			"song_title "\
			"FROM songs "\
			"INNER JOIN pools ON songs.pool_id = pools.pool_id "\
			"WHERE pool_online = 'Y' "\
			"ORDER BY RAND() LIMIT 1";
		}

		D(dolog(LOG_DEBUG, "Query: %s\n", q);)
		if (mysql_query(radio.db, q) != 0)
		{
			dolog(LOG_ERR, "SQL error: %s : %s\n", mysql_error(radio.db), q);
			exit(-1);
		}

		if (	mysql_field_count(radio.db) > 0 &&
			(res = mysql_store_result(radio.db)))
		{
			row = mysql_fetch_row(res);
			if (row)
			{
				char *song_id = strdup(row[1]);
				char *filename = strdup(row[2]);
				char *artist = strdup(row[3]);
				char *title = strdup(row[4]);

				// Free it
				mysql_free_result(res);

				// Delete this vote
				query(	strlen(song_id),
					"UPDATE votes SET vote_handled = 'Y' WHERE song_id = '%s'",
					song_id);
	
				if ((FileSocket = open(row[2], O_LARGEFILE)))
				{
					SetPlayingTitle(song_id, artist, title);
				}
				else
				{
					dolog(LOG_ERR, "Couldn't open %s: %m\n", filename);
					FileSocket = -1;
				}
				free(song_id);
				free(filename);
				free(artist);
				free(title);
			}
			else
			{
				mysql_free_result(res);
			}
			res = NULL;
		}

		if (FileSocket == -1)
		{
			if (!random && row == NULL)
			{
				// No more left in vote queue
				random = 1;
			}
			else
			{
				dolog(LOG_INFO, "Couldn't select a random song\n");
				sleep(10);
			}
		}
	}
}

void GetNewStream()
{
	GetStream(0);
	if (FileSocket == -1) GetStream(1);

	SendTitle();
}

int InitMultiSocket(	int *sock,
			struct addrinfo **resptr,
			const char *szAddr,
			const char *szLocalAddr,
			char *Port,
			int ttl,
			int if_index)
{
	int ret, iLoop = 1;
	char cttl = (char)ttl;
	char cLoop = 1;

	struct addrinfo hints, *res;
	int             error;

	if ((*sock) > 0) return (1);

	*resptr = NULL;

	memset(&hints, 0, sizeof(hints));
	/* set-up hints structure */
	hints.ai_family = PF_UNSPEC;
	hints.ai_flags = AI_PASSIVE;
	hints.ai_socktype = SOCK_DGRAM;
	error = getaddrinfo(szAddr, Port, &hints, &res);
	if (error != 0)
	{
		dolog(LOG_ERR, "Couldn't do getaddrinfo() #1 for %s: %s\n", szAddr, gai_strerror(error));
		return(0);
	}
	/*
	* "res" has a chain of addrinfo structure filled with
	* 0.0.0.0 (for IPv4), 0:0:0:0:0:0:0:0 (for IPv6) and alike,
	* with port filled for "myservice".
	*/
	while (*sock < 0 && res)
	{
		*sock = socket(res->ai_family,
			res->ai_socktype,
			res->ai_protocol);
		if (*sock >= 0)
		{
			dolog(LOG_INFO, "Allocated an %s (%d) socket for %s port %s.\n",
				res->ai_family == AF_INET6 ? "IPv6" :
				res->ai_family == AF_INET ? "IPv4" : "???",
				res->ai_family,
				szAddr,
				Port);
			*resptr = res;
		}
		res = res->ai_next;
	}

	if ((*sock) < 0)
	{
		dolog(LOG_ERR, "UDP socket() failed: %m\n");
		freeaddrinfo(res);
		*resptr = NULL;
		return 0;
	}

	// We need this for some of the options.
	res = *resptr;

	ret = setsockopt((*sock), SOL_SOCKET, SO_REUSEADDR, &iLoop, sizeof(int));
	if (ret < 0)
	{
		dolog(LOG_ERR, "setsockopt SO_REUSEADDR failed: %m\n");
		return 0;
	}

	if (res->ai_family == AF_INET)
	{
		ret = setsockopt((*sock), IPPROTO_IP, IP_MULTICAST_TTL, &cttl, sizeof(char));
	}
	else
	{
		dolog(LOG_INFO, "Setting TTL to %d\n", ttl);
		ret = setsockopt((*sock), IPPROTO_IPV6, IPV6_MULTICAST_HOPS, &ttl, sizeof(ttl));
	}
	if (ret < 0)
	{
		dolog(LOG_ERR, "setsockopt MULTICAST_TTL failed: %m\n");
		return 0;
	}

	if (res->ai_family == AF_INET)
		ret = setsockopt((*sock), IPPROTO_IP, IP_MULTICAST_LOOP, &cLoop, sizeof(char));
	else
		ret = setsockopt((*sock), IPPROTO_IPV6, IPV6_MULTICAST_LOOP, &cLoop, sizeof(char));
	if (ret < 0)
	{
		dolog(LOG_ERR, "setsockopt MULTICAST_LOOP failed: %m\n");
		return 0;
	}

	/* If we have a local addr, bind to that specific address */
	if (szLocalAddr && strlen(szLocalAddr) > 0)
	{
		dolog(LOG_DEBUG, "Trying to set Multicast IP to %s\n", szLocalAddr);
		memset(&hints, 0, sizeof(hints));
		/* set-up hints structure */
		hints.ai_family = PF_INET6;
		error = getaddrinfo(szLocalAddr, NULL, &hints, &res);
		if (error != 0)
		{
			dolog(LOG_ERR, "Couldn't do getaddrinfo() #2 for %s: %s\n", szLocalAddr, gai_strerror(error));
			return(0);
		}
		/*
		* "res" has a chain of addrinfo structure filled with
		* 0.0.0.0 (for IPv4), 0:0:0:0:0:0:0:0 (for IPv6).
		*/
		while (res)
		{
			char buf[200];
			memset(&buf, 0, sizeof(buf));
			inet_ntop(res->ai_family, res->ai_addr->sa_data+6, buf, sizeof(buf));
			ret = bind((*sock),
				res->ai_addr, res->ai_addrlen);
			if (ret < 0)
			{
				dolog(LOG_ERR, "binding to specific %s %s failed: %m\n",
					res->ai_family == AF_INET ? "IPv4" : "IPv6",
					buf);
				return 0;
			}
			else
			{
				dolog(LOG_INFO, "binding to %s %s: ok\n",
					res->ai_family == AF_INET ? "IPv4" : "IPv6",
					buf);
				break;
			}

			res = res->ai_next;
		}
	}

	if (if_index != 0)
	{
		ret = setsockopt((*sock),
			res->ai_family == AF_INET ? IPPROTO_IP : IPPROTO_IPV6,
			res->ai_family == AF_INET ? IP_MULTICAST_IF : IPV6_MULTICAST_IF,
			&if_index, sizeof(if_index));
		if (ret < 0)
		{
			dolog(LOG_ERR, "binding UDP to specific IP failed: %m\n"); return 0;
		}
		else dolog(LOG_INFO, "binding UDP to Interface %d: ok\n", if_index);
	}

	return(1);
}


long long GetTime_ll(void)
{
	struct timeval sTime;

	gettimeofday(&sTime, NULL);
	if (iBaseTime == 0) iBaseTime= sTime.tv_sec;

	return (((long long)sTime.tv_sec - (long long)iBaseTime) * 100000) +
		((long long)sTime.tv_usec / 10);
}

void ResetTime(void)
{
	struct timeval sTime;
	gettimeofday(&sTime, NULL);
	iBaseTime = sTime.tv_sec;
}

int bitrate_table[2][4][16] = {
	/* MPEG II */
	{
		/* reserved */
		{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
		/* layer 3 */
		{0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, 0},
		/* layer 2 */
		{0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, 0},
		/* layer 1 */
		{0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 356, 0}
	},
	/* MPEG I */
	{
		/* reserved */
		{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
		/* layer 3 */
		{0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 0},
		/* layer 2 */
		{0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384, 0},
		/* layer 1 */
		{0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, 0}
	}
};

int samplerate_table[2][3] = {
	{22050, 24000, 16000},
	{44100, 48000, 32000}
};


/*
* Read and decode the first 3 octets from an MP3 frame to work out frame
* size, bitrate and samplerate. If dontread == 1, use the previous values.
* Returns 1 if ok, 0 if an error. Copy the 3 bytes into pBuffer if ok.
*/
static unsigned char fheader[3];

int GetMPEGFrameInfo_b(	unsigned char *pBuffer,
			int dontread,
			int *iFrameSize,
			int *iBitrate,
			int *iSamplerate)
{
	int err;
	int mpeg;
	int layer;
	int bitrate_index;
	int samplerate_index;
	int pad;
	int no_valid_header=1;

	while (no_valid_header)
	{
		if (!dontread)
		{
			fheader[0] = 0;
			while (1)
			{
				err = Read(FileSocket, &fheader[1], 1);
				if (err != 1)
				{
					return (0);
				}
				if (	(unsigned char)fheader[0] == 0xff &&
			    		((unsigned char)fheader[1] & 0xf0) == 0xf0)
				  	break;
				fheader[0] = fheader[1];
			}
			err = Read(FileSocket, &fheader[2], 1);
			if (err != 1)
			{
				return (0);
			}
		}
		mpeg		= (fheader[1] & 0x08) >> 3;
		layer		= (fheader[1] & 0x06) >> 1;
		bitrate_index	= (fheader[2] & 0xf0) >> 4;
		samplerate_index= (fheader[2] & 0x0c) >> 2;
		pad		= (fheader[2] & 0x02) >> 1;
	
		// FIXME: Check bounds.
		*iBitrate	= bitrate_table[mpeg][layer][bitrate_index];
		*iSamplerate	= samplerate_table[mpeg][samplerate_index];

		/* Exit the outside loop if we now have a valid header */
		if (	*iBitrate != 0 &&
			*iSamplerate != 0) break;
	
		dolog(LOG_ERR, "Frame err: mpeg %d layer %d bit %d sample %d pad %d\n",
			mpeg, layer, bitrate_index, samplerate_index, pad);
		dontread = 0; 
	}

	*iFrameSize= (144000 * *iBitrate / *iSamplerate) + pad;

	pBuffer[0] = fheader[0];
	pBuffer[1] = fheader[1];
	pBuffer[2] = fheader[2];
	return (1);
}

int leftover_frame = 0;
/*
* Construct a packet by reading in as many MP3 frames as we can until there
* is no more room in the buf. Return the number of frames read in. If we
* return 0, then file is at the end.
*/
int MakePacket(	unsigned char	*buf,
		const int	pktsize,
		int		*iBytesInPacket,
		int		*iSampleRate)
{
	int iFrameSize, iNumFrames= 0, iBitRate;
	int err;
	int numread;
	
	*iBytesInPacket= 0;
	while (1)
	{
		err = GetMPEGFrameInfo_b(buf, leftover_frame, &iFrameSize, &iBitRate, iSampleRate);
		if (err == 0)
		{
			leftover_frame = 0;
			return (0);
		}
		if ((*iBytesInPacket + iFrameSize) > pktsize)
		{
//			dolog(LOG_ERR, "Can't fit %d and %d into %d\n",
//				*iBytesInPacket, iFrameSize, pktsize);
			leftover_frame = 1;
			return iNumFrames;
		}

		/* We have enough room to read in the rest of the frame */
		leftover_frame = 0;
		(*iBytesInPacket) += iFrameSize;
		buf += 3;
		iFrameSize -= 3;
		while (iFrameSize > 0)
		{
			numread = Read(FileSocket, buf, iFrameSize);
			if (numread <= 0)
			{
				/* A real read failure, give up on the read */
				leftover_frame = 0;
				return 0;
			}
			iFrameSize -= numread;
			buf += numread;
		}
		iNumFrames++;
	}
	dolog(LOG_ERR, "How did we get here with %d frames?\n", iNumFrames);
	return (iNumFrames);
}

/*
 * Spread those mp3's!!! :)
*/
void Work(void)
{
	unsigned char		pBuffer[4096];
	RTPHeader		*pPacket		= (RTPHeader *)pBuffer;
	unsigned int		uPacketNum		= 0,
				time_between_packets	= 0;
	int			packet_delay		= 0;
	long long		llPacketTime		= 0,
				llCurrentTime		= 0;
	unsigned		uSequenceNumber		= 0;
	int			ret			= 0;
	unsigned int		i			= 0;
	int			iSampleRate		= 0;
	int			iBytesCopied		= 0;
	int			iFlagMask		= 0x800E << 16;
	int			frames_in_pkt		= 0;
	int			hdrlen			= sizeof(RTPHeader);
	struct line		*line;

	srand(getpid() + time(NULL));
	memset(pPacket, 0, sizeof(pPacket));
	pPacket->syncsource_id	= htonl(rand());
	pPacket->audioheader	= 0;
	packet_delay		= 0;

	// Reset some 'static' values, as these could be
	// set by a previous Work() and we don't want them
	// to interfere with the new run.
	leftover_frame		= 0;
	memset(fheader, 0, sizeof(fheader));

	for (uPacketNum = 0, uSequenceNumber = rand();; uPacketNum++, uSequenceNumber++)
	{
		/* Make a packet. If we can 't make a packet for some reason, give up */
		frames_in_pkt = MakePacket(&pBuffer[hdrlen],
			radio.c_MTU - sizeof(RTPHeader), &iBytesCopied, &iSampleRate);

		// Couldn't make a packet? -> next song!
		if (frames_in_pkt <= 0 || iSampleRate < 100) break;

		pPacket->flags = htonl(iFlagMask | (uSequenceNumber & 0xFFFF));

		/* iTimeBetweenPackets is in 10us units */
		time_between_packets= ((1000 * 1152) / (iSampleRate / 100)) * frames_in_pkt;

		/* uPacketTime is in 10us units */
		llPacketTime += (long long)time_between_packets;
		pPacket->timestamp= 0;
		
		if (uPacketNum == 0) ResetTime();
	
		llCurrentTime = GetTime_ll();
		packet_delay = (int)(llPacketTime - llCurrentTime);
		if (packet_delay < 0) packet_delay= 0;

		usleep(packet_delay * 10);

		for (i = 0; i < radio.interfaces; i++)
		{
			line = &radio.lines[i];
			ret = sendto(line->Socket, pBuffer, iBytesCopied + sizeof(RTPHeader), 0,
				(struct sockaddr *)line->sSockAddr->ai_addr, line->sSockAddr->ai_addrlen);
			if (ret < 0)
			{
				dolog(LOG_ERR, "sendto() returned: %m\n");
			}
		}
  	}
	return;
}

// Next song :)
void sighup(int i)
{
	close(FileSocket);
	FileSocket = -1;
	signal(SIGHUP, &sighup);
}

void savepid()
{
	FILE *f = fopen(radio.c_PIDFile, "w");
	if (!f) return;
	fprintf(f, "%d", getpid());
	fclose(f);
}

void cleanpid(int i)
{
	unlink(radio.c_PIDFile);
	exit(0);
}

int main(int argc, char *argv[])
{
	unsigned int	i;
	int		err;
	char		*sError;
	struct line	*line;

	memset(&radio, 0, sizeof(radio));
	
	/* Read config file */
	if (!ReadConfigFile())
	{
		dolog(LOG_ERR, "Bad or missing config file, giving up\n");
		return -1;
	}

	/* Report the settings */
        dolog(LOG_INFO, "StreamTitle          : %s\n", radio.c_StreamTitle);
	dolog(LOG_INFO, "URL                  : %s\n", radio.c_URL);
	if (radio.c_Multicast_Source)
	{
		dolog(LOG_INFO, "Multicast Source     : %s\n", radio.c_Multicast_Source);
	}
	dolog(LOG_INFO, "Multicast Address    : %s\n", radio.c_Multicast_Address);
	dolog(LOG_INFO, "Multicast Audio Port : %s\n", radio.c_Multicast_Audio_Port);
	dolog(LOG_INFO, "Multicast Title Port : %s\n", radio.c_Multicast_Title_Port);
	dolog(LOG_INFO, "Database Server      : %s\n", radio.c_DB_Server);
	dolog(LOG_INFO, "Database Name        : %s\n", radio.c_DB_Name);
	dolog(LOG_INFO, "Database User        : %s\n", radio.c_DB_User);
	dolog(LOG_INFO, "Database Password    : %s\n", strlen(radio.c_DB_Password) > 0 ? "***" : "");
	dolog(LOG_INFO, "Interfaces           : %u\n", radio.interfaces);

	radio.db = mysql_init(NULL);
	if (!radio.db)
	{ 
		dolog(LOG_ERR, "Couldn't initialize MySQL\n");
		return -1;
	}

	if (!mysql_real_connect(
		radio.db,
		radio.c_DB_Server,
		radio.c_DB_Name,
		radio.c_DB_User,
		radio.c_DB_Password,
		0, NULL, 0))
	{
		dolog(LOG_ERR, "Couldn't open database\n");
		return -1;
	}

	for (i=0;i<radio.interfaces;i++)
	{
		line = &radio.lines[i];
		/* Make the title output socket */
		err = InitMultiSocket(
			&line->TitleSocket,
			&line->sTitleAddr,
			radio.c_Multicast_Address,
			radio.c_Multicast_Source,
			radio.c_Multicast_Title_Port,
			radio.c_HopLimit,
			line->Interface);
		if (err != 1)
		{
			dolog(LOG_ERR, "Couldn't create title socket, exiting\n");
			return -1;
		}

		/* We can now create our RTP output sockets */
		err = InitMultiSocket(
			&line->Socket,
			&line->sSockAddr,
			radio.c_Multicast_Address,
			radio.c_Multicast_Source,
			radio.c_Multicast_Audio_Port,
			radio.c_HopLimit,
			line->Interface);
		if (err != 1)
		{
			dolog(LOG_ERR, "Couldn't create RTP socket1, exiting\n");
			return -1;
		}
	}
#ifndef DEBUG
	if (radio.c_Daemonize)
	{
		// Daemonize
		err  = fork();
		if (err < 0)
		{
			fprintf(stderr, "Couldn't fork\n");
			return -1;
		}
		// Exit the mother fork
		if (err != 0) return 0;

		// Child fork
		setsid();
		// Cleanup stdin/out/err
		freopen("/dev/null","r",stdin);
		freopen("/dev/null","w",stdout);
		freopen("/dev/null","w",stderr);
	}
#endif
	savepid();

	// SIGHUP causes the song to be skipped
	signal(SIGHUP, &sighup);

	// SIGKILL/TERM/INT causes the pidfile to be removed
	signal(SIGKILL, &cleanpid);
	signal(SIGTERM, &cleanpid);
	signal(SIGINT, &cleanpid);

	// Show our version in the startup logs
	dolog(LOG_INFO, "Radio.Unfix %s by Jeroen Massar <jeroen@unfix.org>\n", RADIO_VERSION);

	// Do the work
	while (1)
	{
		GetNewStream();
		if (FileSocket != -1)
		{
			Work();
			if (FileSocket != -1)
			{
				close(FileSocket);
				FileSocket = -1;
			}
		}
		else
		{
			dolog(LOG_INFO, "Nothing to play... waiting 10 seconds\n");
			sleep(10);
		}
	}

	dolog(LOG_INFO, "Shutdown, thank you for using Radio.Unfix\n");

	// Close the database
	mysql_close(radio.db);

	cleanpid(SIGINT);
	return 0;
}

